package Atlas::DB::Import;

use strict;
use warnings;
use Carp qw(croak carp);

use DBI;
use Log::Log4perl qw(get_logger :levels);
use Data::Dumper;
use Readonly;
use List::Compare;
use Perl6::Slurp;

use File::Temp;

use Digest::MD5 qw(md5_hex);

sub new {
    my $class = shift;
    my %args  = (
        @_,
    );

    bless \%args, $class;
}

sub init_log {
    my $self = shift;
    my %args = (
        @_,
    );

    if ((exists $self->{'log4perl-config-file'}) && (-f $self->{'log4perl-config-file'})) {
        Log::Log4perl::init_and_watch($self->{'log4perl-config-file'}, $self->{'log4perl-config-file-watch-time'});
    }
    else {
        my $logger = get_logger(__PACKAGE__);
        $logger->level($INFO);

        my $appender = Log::Log4perl::Appender->new(
            'Log::Dispatch::Syslog',
        );
        $logger->add_appender($appender);

        my $layout = Log::Log4perl::Layout::PatternLayout->new("%d %p> %F{1}:%L %M - %m%n");
        $appender->layout($layout);
    }
}

sub init_db {
    my $self = shift;
    my %args = (
        shall_die => 0,
        @_,
    );

    # Set up some constants. They are somewhat arbitrary.
    Readonly my $ENCRYPTED_PASSWORD           => '*';     # This seems to mean that there is no password hash
    Readonly my $DATE_OF_LAST_PASSWORD_CHANGE => 14856;   # A meaningless constant.
    Readonly my $MINIMUM_PASSWORD_AGE         => 0;       # We don't need to enforce a minimum password age.
    Readonly my $MAXIMUM_PASSWORD_AGE         => 99999;   # Make the password last as long as we can.
    Readonly my $PASSWORD_WARNING_PERIOD      => 7;       # One week warning period.
    Readonly my $PASSWORD_INACTIVITY_PERIOD   => 99999;   # No password inactivity period.
    Readonly my $ACCOUNT_EXPIRATION_DATE      => 99999;   # Don't expire according to this table.
    Readonly my $RESERVED_FIELD               => 0;       # Unused at this point. 0 is as good a value as anything.

    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    INITIALIZE_DATABASE:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;

        undef($self->{source_dbh});
        undef($self->{destination_dbh});

        my $source_database_directive      = ($self->{'source-database-driver'}      eq 'Pg') ? 'dbname' : 'database';
        my $destination_database_directive = ($self->{'destination-database-driver'} eq 'Pg') ? 'dbname' : 'database';

        my $source_dsn      = "DBI:$self->{'source-database-driver'}:$source_database_directive=$self->{'source-database-name'};host=$self->{'source-database-host'}";
        my $destination_dsn = "DBI:$self->{'destination-database-driver'}:$destination_database_directive=$self->{'destination-database-name'};host=$self->{'destination-database-host'}";

        $self->{'source_dbh'}      = DBI->connect($source_dsn,      $self->{'source-database-user'},      $self->{'source-database-pass'},      { RaiseError => 0, AutoCommit => 1 });
        $self->{'destination_dbh'} = DBI->connect($destination_dsn, $self->{'destination-database-user'}, $self->{'destination-database-pass'}, { RaiseError => 1, AutoCommit => 1 });

        # Verify connection to source DB
        if (! $self->{source_dbh}) {
            my $err = "Could not connect to $source_dsn. (".$DBI::errstr.")";
            $logger->error($err);
            croak $err if ($args{shall_die});
            if ($DBI::err == 1 || $DBI::err == 7) {
                next INITIALIZE_DATABASE;
            }
            else {
                $logger->logcroak('DBI died. We would normally restart, but we need the code. DBI err code: '.$DBI::err);
            }
        }

        # Verify connection to destination DB
        if (! $self->{destination_dbh}) {
            my $err = "Could not connect to $destination_dsn. (".$DBI::errstr.")";
            $logger->error($err);
            croak $err if ($args{shall_die});
            if ($DBI::err == 1 || $DBI::err == 7) {
                next INITIALIZE_DATABASE;
            }
            else {
                $logger->logcroak('DBI died. We would normally restart, but we need the code. DBI err code: '.$DBI::err);
            }
        }

        # Setup Statement Handles (sth). If those fail to be setup, "next" has the loop
        # forget about the DB connections and try again.
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_source_accounts',
            database_source => 'source',
            sth_sql         => $self->{'source-query'},
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'sanity_check_query',
            database_source => 'source',
            sth_sql         => $self->{'sanity-check-query'},
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_destination_accounts',
            database_source => 'destination',
            sth_sql         => 'SELECT username, comment, id AS unix_id, home_directory FROM users',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'insert_into_destination',
            database_source => 'destination',
            sth_sql         => 'INSERT INTO users '.
                               '(username, password, id, group_id, comment, home_directory, encrypted_password,    date_of_last_password_change,  minimum_password_age,  maximum_password_age,  password_warning_period,  password_inactivity_period,  account_expiration_date,  reserved_field) '.
                               'VALUES '.
                               "(?,        ?,        ?,  ?,        ?,       ?,              '$ENCRYPTED_PASSWORD', $DATE_OF_LAST_PASSWORD_CHANGE, $MINIMUM_PASSWORD_AGE, $MAXIMUM_PASSWORD_AGE, $PASSWORD_WARNING_PERIOD, $PASSWORD_INACTIVITY_PERIOD, $ACCOUNT_EXPIRATION_DATE, $RESERVED_FIELD)"
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'update_destination_user',
            database_source => 'destination',
            sth_sql         => 'UPDATE users SET comment = ? WHERE id = ?',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'rename_destination_user',
            database_source => 'destination',
            sth_sql         => 'UPDATE users SET username = ?, comment = ?, home_directory = ? WHERE id = ?',
        );
        $done = 1;
    }
}

sub init_sth {
    my $self = shift;
    my %args = (
        sql_key         => undef,
        sth_sql         => undef,
        shall_die       => 0,
        database_source => undef, # source || destination
        @_,
    );
    my $db_src = $args{'database_source'};

    my $logger = get_logger(__PACKAGE__);

    my $sth_key = $args{sql_key}.'_sth';
    undef($self->{$sth_key});

    $self->{$args{sql_key}} = $args{sth_sql};

    $self->{$sth_key} = $self->{$db_src . '_dbh'}->prepare($args{sth_sql});
    if (! $self->{$sth_key}) {
        my $err = "Database error: ".$self->{$db_src . 'dbh'}->errstr." (while preparing '$args{sth_sql}')";
        $logger->error($err);
        croak $err if ($args{shall_die});
        if ($DBI::err == 1 || $DBI::err == 7) {
            return 0; # Return of 0 means to try again.
        }
        else {
            $logger->logcroak("DBI died. We would normally restart, but we need the code. DBI err code: ".$DBI::err);
        }
    }

    return 1; # Success
}

sub sanity_check {
    my $self = shift;
    my %args = (
        @_,
    );

    my $sql = 'sanity_check_query';
    my $sth = $sql.'_sth';

    my $sanity_check_result = undef;

    my $query_log_statement = 'Query of ['.$self->{$sql}.']';
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_SANITY_CHECK_RESULT:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $sanity_check_result = $self->{$sth}->fetchall_arrayref;
            if ($self->{$sth}->err) {
                my $err = 'Database error: '.$self->{$sth}->errstr
                        . ' and database error code of '.$self->{$sth}->err
                        . " (while fetchall_arrayref-ing '$query_log_statement')"
                        ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err  .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_SANITY_CHECK_RESULT;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while executing '$query_log_statement')"
                    ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err  .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next GET_SANITY_CHECK_RESULT;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    if (@$sanity_check_result != 0) {
        $logger->logcroak("Sanity check failed. Query $self->{$sql} returned non-zero result set. Expecting no rows but ".scalar(@$sanity_check_result)." returned.");
    }

    return;
}

=head2
get_source_users - Get array of hashes of source users.
=cut
sub get_source_users {
    my $self = shift;
    my %args = (
        @_,
    );

    Readonly my $SOURCE_USER_KEY => 'username';

    my $sql = 'select_source_accounts';
    my $sth = $sql.'_sth';

    my $source_users = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";# with parameters of (".join(', ', @params).")";
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_SOURCE_USERS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $source_users = $self->{$sth}->fetchall_hashref($SOURCE_USER_KEY);
            if ($self->{$sth}->err) {
                my $err  = 'Database error: '.$self->{$sth}->errstr
                         . ' and database error code of '.$self->{$sth}->err
                         . " (while fetchall_hashref-ing '$query_log_statement')"
                         ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err  .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_SOURCE_USERS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err  = 'Database error: '.$self->{$sth}->errstr
                     . ' and database error code of '.$self->{$sth}->err
                     . " (while executing '$query_log_statement')"
                     ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err  .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_SOURCE_USERS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    $self->{source_users} = $source_users;

    my %source_unix_ids_hash = map { $source_users->{$_}{unix_id} => $source_users->{$_} } keys %$source_users;
    $self->{source_unix_ids} = \%source_unix_ids_hash;

    $self->{source_users_count} = scalar keys %$source_users;

    return;
}

=head2
get_destination_users - Get array of hashes of destionation users.
=cut

sub get_destination_users {
    my $self = shift;
    my %args = (
        @_,
    );

    Readonly my $DESTINATION_USER_KEY => 'username';

    my $sql = 'select_destination_accounts';
    my $sth = $sql.'_sth';

    my $destination_users = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";# with parameters of (".join(', ', @params).")";
    my $logger = get_logger(__PACKAGE__);
    my $done = -1;

    GET_DESTINATION_USERS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;

        if ($self->{$sth}->execute) {
            $destination_users = $self->{$sth}->fetchall_hashref($DESTINATION_USER_KEY);
            if ($self->{$sth}->err) {
                my $err  = 'Database error: '.$self->{$sth}->errstr
                         . ' and database error code of '.$self->{$sth}->err
                         . " (while fetchall_arrayref-ing '$query_log_statement')"
                         ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err  .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_DESTINATION_USERS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err  = 'Database error: '.$self->{$sth}->errstr
                     . ' and database error code of '.$self->{$sth}->err
                     . " (while executing '$query_log_statement')"
                     ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err  .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_DESTINATION_USERS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    $self->{destination_users} = $destination_users;

    my %destination_unix_ids_hash = map { $destination_users->{$_}{unix_id} => $destination_users->{$_} } keys %$destination_users;
    $self->{destination_unix_ids} = \%destination_unix_ids_hash;

    return;
}

sub check_consistency {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    Readonly my $DEFAULT_PASSWORD => 'x';
    Readonly my $DEFAULT_GROUP_ID => 1003;  # This maps to the "external" group.

    my $users_processed_count = 0;
    PROCESS_USER:
    while (my ($source_username, $source_record) = each %{$self->{source_users}}) {

        if ($source_username !~ m{
            \A     # beginning of string
            [\w-]+ # username must have only alphanumeric characters or a dash
            \z     # end of string
        }xms) {
            $logger->logcarp("Warning! Source user $source_username did not match the safety check, skipping.");
            next PROCESS_USER;
        }

        my $source_unix_id = $source_record->{unix_id};
        if (($users_processed_count % 1000) == 1) {
            if (! $self->{quiet}) {
                my $percent_complete = sprintf("%.3f", ($users_processed_count / $self->{source_users_count} * 100));
                print "$percent_complete% Complete\n";
            }
        }

        $logger->debug("Source record: ", { filter => \&Data::Dumper::Dumper, value => $source_record });
        $logger->debug("Destination record for username $source_username: ", { filter => \&Data::Dumper::Dumper, value => $self->{destination_users}{$source_username} });
        $logger->debug("Destination record for unix id $source_unix_id: ", { filter => \&Data::Dumper::Dumper, value => $self->{destination_unix_ids}{$source_unix_id} });

        if (! exists $self->{destination_users}{$source_username} && ! exists $self->{destination_unix_ids}{$source_unix_id}) {
            $logger->debug("User $source_username, nor does UNIX ID $source_unix_id, exist in destination database.");
            $logger->info("Inserting $source_username into destination database.");
            $self->insert_user(
                username       => $source_username,
                password       => $DEFAULT_PASSWORD,
                unix_id        => $source_record->{unix_id},
                group_id       => $DEFAULT_GROUP_ID,
                comment        => $source_record->{comment},
                home_directory => $self->compute_home_directory(username => $source_username),
            );
            $source_record->{home_directory} = $self->compute_home_directory(username => $source_username);
            $self->perform_per_entry_command_generation(type => 'insert', record => $source_record);
        }
        elsif (exists $self->{destination_users}{$source_username} && exists $self->{destination_unix_ids}{$source_unix_id}) {
            if ($self->{destination_users}{$source_username}{unix_id} == $source_unix_id) {
                $logger->debug("Checking username $source_username fields in source and destination to determine if we need to update.");

                # List of fields that we will check to determine if we need to update the destination database with data from the
                # source database.
                my @significant_fields = qw{ comment };

                # Use hash slicing to get data from source record.
                my @source_data = $source_record->{@significant_fields};

                # Again, we need the value of the above "significant" fields for the destination database record. Use hash slicing.
                my @destination_data = $self->{destination_unix_ids}{$source_unix_id}{@significant_fields};

                # Make a list compare object.
                my $list_compare = List::Compare->new('--unsorted', \@source_data, \@destination_data);

                # Test the size of the intersection between source and destination with size of the significant fields.
                if (scalar($list_compare->get_intersection) != scalar(@significant_fields)) {;
                    # Then we can update.
                    $logger->info("Updating $source_username in destination database.");
                    $self->update_user(
                        username       => $source_username,
                        comment        => $source_record->{comment},
                    );
                }
            }
            else {
                $logger->logcroak(
                    "Source username, unix id mismatch to destination username, unix id. "
                    . "Source username: $source_username source unix id: $source_unix_id. "
                    . "Destination username: $source_username destination unix id: $self->{destination_users}{$source_username}{unix_id}."
                    );
            }
        }
        elsif (! exists $self->{destination_users}{$source_username} && exists $self->{destination_unix_ids}{$source_unix_id}) {
            # UNIX ID exists, but username does not. We assume that this is a rename
            $logger->info("Renaming $self->{destination_unix_ids}{$source_unix_id}{username} to $source_username in destination database.");
            my $new_home_directory = $self->compute_home_directory(username => $source_username);
            my $old_home_directory = $self->{destination_unix_ids}{$source_unix_id}{home_directory};
            $self->rename_user(
                username       => $source_username,
                comment        => $source_record->{comment},
                home_directory => $new_home_directory,
                unix_id        => $source_record->{unix_id},
            );
            $source_record->{new_home_directory} = $new_home_directory;
            $source_record->{old_home_directory} = $old_home_directory;
            $self->perform_per_entry_command_generation(type => 'rename', record => $source_record);
        }
        elsif (exists $self->{destination_users}{$source_username} && ! exists $self->{destination_unix_ids}{$source_unix_id}) {
            $logger->logcroak("Source username $source_username is found in destination database, but source unix id $source_unix_id is not.");
        }
        else {
            $logger->logcroak('Unhandled condition. We should never reach this portion of the code.');
        }

        $users_processed_count++;
    }
}

sub compute_home_directory {
    my $self = shift;
    my %args = (
        @_,
    );
    my $logger = get_logger(__PACKAGE__);

    # Take the low order byte from the md5sum - we also want to ensure it is lowercase.
    my $map = lc substr md5_hex($args{username}), -2;

    return "/home/$map/$args{username}";
}

sub insert_user {
    my $self = shift;
    my %args = (
        @_,
    );
    my $logger = get_logger(__PACKAGE__);

    my $sql = 'insert_into_destination';
    my $sth = $sql.'_sth';

    my @params = (
        $args{username},
        $args{password},
        $args{unix_id},
        $args{group_id},
        $args{comment},
        $args{home_directory},
    );
    $logger->debug("params for insert_user sql:", { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of (".join(', ', @params).")";

    my $done = -1;

    INSERT_USER:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq "0E0") {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err  = 'Database error: '.$self->{$sth}->errstr
                     . ' and database error code of '.$self->{$sth}->err
                     . " (while executing '$query_log_statement')"
                     ;
            if ($self->{$sth}->err == 1 or $self->{$sth}->err == 7) {
                $err  .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next INSERT_USER;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub update_user {
    my $self = shift;
    my %args = (
        @_,
    );
    my $logger = get_logger(__PACKAGE__);

    my $sql = 'update_destination_user';
    my $sth = $sql.'_sth';

    my @params = (
        $args{comment},
        $args{unix_id},
    );
    $logger->debug('params for update_user sql:', { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = 'Query of ['.$self->{$sql}.'] with parameters of ('.join(', ', @params).')';

    my $done = -1;

    UPDATE_USER:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq "0E0") {
                $logger->error("Updating user, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Updating user, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err  = 'Database error: '.$self->{$sth}->errstr
                     . ' and database error code of '.$self->{$sth}->err
                     . " (while executing '$query_log_statement')"
                     ;
            if ($self->{$sth}->err == 1 or $self->{$sth}->err == 7) {
                $err  .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next UPDATE_USER;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub rename_user {
    my $self = shift;
    my %args = (
        @_,
    );
    my $logger = get_logger(__PACKAGE__);

    my $sql = 'rename_destination_user';
    my $sth = $sql.'_sth';

    my @params = (
        $args{username},
        $args{comment},
        $args{home_directory},
        $args{unix_id},
    );
    $logger->debug('params for rename_user sql:', { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = 'Query of ['.$self->{$sql}.'] with parameters of ('.join(', ', @params).')';

    my $done = -1;

    RENAME_USER:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {
            $rows = $self->{$sth}->execute(@params);
        };
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq "0E0") {
                $logger->error("Renaming user, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Renaming user, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err  = 'Database error: '.$self->{$sth}->errstr
                     . ' and database error code of '.$self->{$sth}->err
                     . " (while executing '$query_log_statement')"
                     ;
            if ($self->{$sth}->err == 1 or $self->{$sth}->err == 7) {
                $err  .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next RENAME_USER;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub perform_per_entry_command_generation {
    my $self = shift;
    my %args = (
        type   => 'insert',
        record => undef,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    if (exists $self->{'per-entry-commands'}{$args{type}}) {
        for my $template (@{$self->{'per-entry-commands'}{$args{type}}}) {
            my $fh = File::Temp->new(
                UNLINK => 0,
                DIR    => '/var/lib/atlas/account-importer/per-entry-command.d',
            );
            my $interpolated_template = $self->interpolate_template(
                template => $template,
                record   => $args{record},
                type     => $args{type},
            );
            $logger->debug("Generated template of '$interpolated_template' to file '".$fh->filename."'");
            print { $fh } $interpolated_template or $logger->logcroak("Unable to write template to ".$fh->filename);
            $fh->close or $logger->logcroak("Unable to close temp file ".$fh->filename);
        }
    }
    else {
        $logger->debug("No config entries for per-entry-command for $args{type}.");
    }
    return;
}

sub interpolate_template {
    my $self = shift;
    my %args = (
        template => 'echo <% $record->{username} %>', # A simple example to echo the username
        record   => { username => 'jdoe', },
        type     => 'insert',
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $template = $args{template};
    my $record   = $args{record};

    my $regex = qr{
        <%    # Start of "embedded Perl" section
        \s*   # optional space
        (\S+) # One or more non-space characters
        \s*   # optional space
        %>
    }xms;

    my $done = 0;
    while (!$done) {
        my $replace = undef;
        if (my ($embedded_code) = $template =~ $regex) {
            $replace = eval $embedded_code;
            if ($@) {
                $logger->logcroak("Unable to eval: '$embedded_code' from $args{type} $template.");
            }
            if (defined $replace) {
                if ($template =~ s/$regex/$replace/) {
                    $logger->debug("Successfully replaced $embedded_code with $replace in $args{template}.");
                }
                else {
                    $logger->logcroak("Unable to find $embedded_code in $args{template}.");
                }
            }
            else {
                $logger->logcroak("Unable to find useful value to replace '$embedded_code' with.");
            }
        }
        else {
            $logger->debug('Could not find any more embedded Perl sections.');
            $done = 1;
        }
    }

    return $template;
}

sub process_per_entry_commands {
    my $self = shift;

    my $logger = get_logger(__PACKAGE__);

    # Only perform run-parts if the directory exists
    if (-d $self->{'per-entry-command-pre-directory'}) {
        my $system_command = "run-parts $self->{'per-entry-command-pre-directory'}";
        $self->run_system_command(
            command         => $system_command,
            on_success_ref  => sub {
                my ($logger, $system_command) = @_;
                $logger->info("System command '$system_command' ran successfully.");
            },
            on_success_args => [
                $logger,
                $system_command,
            ],
        );
    }
    else {
        $logger->debug("per-entry-command-pre-directory of $self->{'per-entry-command-pre-directory'} does not exist.");
    }
    # Only continue here if the directory exists
    if (-d $self->{'per-entry-command-directory'}) {
        opendir my $dh, $self->{'per-entry-command-directory'}
            or $logger->logcroak("Unable to open directory $self->{'per-entry-command-directory'} for reading.");
        # Loop through all the files in 'per-entry-command-directory'
        # These files each contain a little shell snippet that help
        # in processing the accounts - things to do when we "insert",
        # "rename", etc. different accounts.
        while (my $relative_file = readdir $dh) {
            my $file = "$self->{'per-entry-command-directory'}/$relative_file";
            # Ensure that the file is regular
            if (-f $file) {
                my $file_text = slurp $file, { chomp => '[EOL]', };
                $logger->info("About to interpret file $file with content: $file_text");

                # And execute the sh(ell) on the file
                $self->run_system_command(
                    command => "sh $file",
                    on_success_ref => sub {
                        my ($logger, $file) = @_;
                        $logger->info("Executing file $file successful. Now unlinking.");
                        unlink $file or $logger->logcroak("Unable to unlink file $file: $!.");
                    },
                    on_success_args => [
                        $logger,
                        $file,
                    ],
                );
            }
        }
        closedir $dh or $logger->logcroak("Unable to close directory $self->{'per-entry-command-directory'}.");
    }
    else {
        $logger->debug("per-entry-command-directory of $self->{'per-entry-command-directory'} does not exist.");
    }
    return;
}

=item run_system_command

Run a "system" command and handle signals, return codes, etc.

Pass the command to run via system and optionally a code reference to run when the system call
exits with a value of 0, a successful exit. You can also pass arguments to the code reference.

=cut
sub run_system_command {
    my $self = shift;
    my %args = (
        command         => undef,
        on_success_ref  => undef,
        on_success_args => [],
        @_,
    );

    return unless defined $args{command};

    my $logger = get_logger(__PACKAGE__);

    my $system_command = $args{command};
    system $system_command;
    if ($? == -1) {
        $logger->logcroak("Unable to run system command '$system_command' $!.");
    }
    elsif ($? & 127) {
        $logger->logcroak(
            sprintf(
                "System command '$system_command' died with signal %d, %s coredump.",
                ($? & 127),
                ($? & 128) ? 'with' : 'without',
            )
        );
    }
    else {
        my $rc = $? >> 8;
        if ($rc == 0) {
            if (defined $args{on_success_ref}) {
                &{$args{on_success_ref}}(@{$args{on_success_args}});
            }
        }
        else {
            $logger->logcroak("System command '$system_command' failed. Returned value of $rc.");
        }
    }
}

sub debug {
    my $self = shift;
    print "Source Users:\n";
    print Dumper($self->{'source_users'});
    print "Destination Users:\n";
    print Dumper($self->{'destination_users'});
    return;
}

1;
