package Atlas::DB::PasswordImport;

use strict;
use warnings;
use Carp qw(croak carp);

use DBI;
use Log::Log4perl qw(get_logger :levels);
use Data::Dumper;
use YAML qw(LoadFile);

use DateTime;
use DateTime::TimeZone;
use Text::Trim;

use Readonly;

sub new {
    my $class = shift;
    my %args  = (
        @_,
    );

    bless \%args, $class;
}

sub init_time_zone {
    my $self = shift;
    my %args = (
        @_,
    );

    $self->{time_zone} = DateTime::TimeZone->new(name => 'local');
}

sub init_log {
    my $self = shift;
    my %args = (
        @_,
    );

    if ((exists $self->{'log4perl-config-file'}) && (-f $self->{'log4perl-config-file'})) {
        Log::Log4perl::init_and_watch($self->{'log4perl-config-file'}, $self->{'log4perl-config-file-watch-time'});
    }
    else {
        my $logger = get_logger(__PACKAGE__);
        $logger->level($INFO);

        my $appender = Log::Log4perl::Appender->new(
            'Log::Dispatch::Syslog',
        );
        $logger->add_appender($appender);

        my $layout = Log::Log4perl::Layout::PatternLayout->new("%d %p> %F{1}:%L %M - %m%n");
        $appender->layout($layout);
    }
}

sub init_db {
    my $self = shift;
    my %args = (
        shall_die => 0,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    Readonly my $INITIALIZE_DATABASE_RETRY_SLEEP_TIME => 5;

    $logger->debug("About to initialize database.");
    my $done = -1;

    INITIALIZE_DATABASE:
    while ($done <= 0) {
        if ($done == 0) {
            $logger->debug("Previous attempt to initialize database failed. Now sleeping for $INITIALIZE_DATABASE_RETRY_SLEEP_TIME.");
            sleep $INITIALIZE_DATABASE_RETRY_SLEEP_TIME;
        }
        $done = 0;

        undef($self->{dbh});

        my $database_directive = ($self->{'database-driver'} eq 'Pg') ? 'dbname' : 'database';

        my $dsn = "DBI:$self->{'database-driver'}:$database_directive=$self->{'database-name'};host=$self->{'database-host'}";
        $logger->debug("About to make database connection with DSN of '$dsn'.");

        $self->{dbh} = DBI->connect($dsn, $self->{'database-user'}, $self->{'database-pass'}, { RaiseError => 0, AutoCommit => 1 });

        # Verify connection to DB
        if (! $self->{dbh}) {
            my $err = "Could not connect to $dsn. (".$DBI::errstr.")";
            $logger->error($err);
            croak $err if ($args{shall_die});
            if ($DBI::err == 1 || $DBI::err == 7) {
                next INITIALIZE_DATABASE;
            }
            else {
                $logger->logcroak('DBI died. We would normally restart, but we need the code. DBI err code: '.$DBI::err);
            }
        }

        # Setup Statement Handles (sth). If those fail to be setup, "next" has the loop
        # forget about the DB connections and try again.
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key => 'select_users',
            sth_sql => 'SELECT username, id AS unix_id, legacy_home_directory FROM users',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key => 'select_passwords',
            sth_sql => 'SELECT * FROM passwords',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key => 'insert_into_passwords',
            sth_sql => 'INSERT INTO passwords '
                       . '(username, password, type, changed_at) '
                       . 'VALUES '
                       . '(?,       ?,        ?,    ?)'
                       ,
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key => 'update_passwords',
            sth_sql => 'UPDATE passwords '
                       . 'SET password = ?, changed_at = ? '
                       . 'WHERE username = ? AND type = ?'
                       ,
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key => 'update_legacy_home_directory',
            sth_sql => 'UPDATE users '
                       . 'SET legacy_home_directory = ? '
                       . 'WHERE username = ?'
                       ,
        );
        $done = 1;
    }
}

sub init_sth {
    my $self = shift;
    my %args = (
        sql_key   => undef,
        sth_sql   => undef,
        shall_die => 0,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    $logger->debug("Initializing statement handle of '$args{sql_key}' with SQL of '$args{sth_sql}'.");

    my $sth_key = $args{sql_key}.'_sth';
    undef($self->{$sth_key});

    $self->{$args{sql_key}} = $args{sth_sql};

    $self->{$sth_key} = $self->{dbh}->prepare($args{sth_sql});
    if (! $self->{$sth_key}) {
        my $err = "Database error: ".$self->{dbh}->errstr." (while preparing '$args{sth_sql}')";
        $logger->error($err);
        croak $err if ($args{shall_die});
        if ($DBI::err == 1 || $DBI::err == 7) {
            return 0; # Return of 0 means to try again.
        }
        else {
            $logger->logcroak('DBI died. We would normally restart, but we need the code. DBI err code: '.$DBI::err);
        }
    }

    return 1; # Success
}

sub acquire_database_records {
    my $self = shift;

    $self->get_users;
    $self->get_passwords;
}

=head2
get_users - Get hash of users.
=cut
sub get_users {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'select_users';
    my $sth = $sql.'_sth';

    my $users = undef;

    my $query_log_statement = 'Query of ['.$self->{$sql}.']';

    my $done = -1;

    GET_USERS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $users = $self->{$sth}->fetchall_hashref('username');
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_hashref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= ' However, found known error code and reinitializing database.';
                    $logger->warn($err);
                    $self->init_db;
                    next GET_USERS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next GET_USERS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    $self->{users} = $users;

    return;
}

=head2
get_passwords - Put passwords into a hash ref keyed by UNIX id.
=cut
sub get_passwords {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'select_passwords';
    my $sth = $sql.'_sth';

    my $passwords = undef;

    my $query_log_statement = 'Query of ['.$self->{$sql}.']';

    my $done = -1;

    GET_PASSWORDS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            while (my $pw = $self->{$sth}->fetchrow_hashref) {
                if (! exists $passwords->{$pw->{username}}) {
                    $passwords->{$pw->{username}} = {};
                }
                $passwords->{$pw->{username}}{$pw->{type}}
                    = {
                        hash       => $pw->{password},
                        changed_at => $self->to_DateTime(timestamp => $pw->{changed_at}),
                    }
                ;
            }
        }
        $done = 1;
    }

    $self->{passwords} = $passwords;

    return;
}

sub reconcile_password_output {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    if (-f $self->{'command-output-map-file'}) {
        my $output_map = LoadFile($self->{'command-output-map-file'});
        if (defined $output_map->{command}) {
            my $output = `$output_map->{command}`;
            if ($? == -1) {
                $logger->logcroak("Unable to run system command '$output_map->{command}' $!.");
            }
            elsif ($? & 127) {
                $logger->logcroak(
                    sprintf(
                        "System command '$output_map->{command}' died with signal %d, %s coredump.",
                        ($? & 127),
                        ($? & 128) ? 'with' : 'without',
                    )
                );
            }
            else {
                my $rc = $? >> 8;
                if ($rc == 0) {
                    $logger->info("System command '$output_map->{command}' executed successfully.");
                }
                else {
                    $logger->logcroak("System command '$output_map->{command}' failed. Returned value of $rc.");
                }
            }
            my @lines = split /\n/, $output;
            my $split_regex = (defined $output_map->{delimiter}) ? $output_map->{delimiter} : '\|';
            my $mapping = $output_map->{mapping};
            PROCESS_OUTPUT_LINE:
            for my $line (@lines) {
                $logger->debug("Processing line: $line");

                # split the data up into the correct chunks
                my @data = split $split_regex, $line;

                # initialize a hash to help in processing
                my %mapped_data;

                # Perform a fancy array slice to assign each key the corresponding value
                @mapped_data{keys %$mapping} = @data[values %$mapping];

                $mapped_data{username} = $self->normalize_username(
                    username => $mapped_data{username},
                );

                $logger->debug('Sliced data: ', { filter => \&Data::Dumper::Dumper, value => \%mapped_data });

                # Sanity check. A value of 1 means we can continue.
                my $sanity_check_result = $self->sanity_check(data => \%mapped_data);
                if (! $sanity_check_result) {
                    $logger->logcarp("Skipping processing line '$line' because santiy check failed.");
                    next PROCESS_OUTPUT_LINE;
                }

                $self->process_legacy_home_directory(data => \%mapped_data);
                $self->process_passwords(
                    data => \%mapped_data,
                    map  => $output_map,
                );
            }
        }
        else {
            $logger->logcroak("command-output-map-file needs a 'command' entry.");
        }
    }
}

sub normalize_username {
    my $self = shift;
    my %args = (
        username => undef,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $retval = $args{username};

    if ($args{username} =~ m{
        \A  # beginning of string
        \s+ # whitespace exists
        |   # or
        \s+ # whitespace exists
        \z  # end of string
    }xms) {
        $retval = trim($args{username});
        $logger->info("Normalizing username ($args{username}) to ($retval).");
    }
    return $retval;
}

sub sanity_check {
    my $self = shift;
    my %args = (
        data => {},
        @_,
    );

    my $data = $args{data};

    my $logger = get_logger(__PACKAGE__);

    if ($data->{username} =~ m{
        \A     # beginning of string
        [\w-]+ # username must have only word characters or a dash
        \z     # end of string
    }xms) {
        $logger->debug("Username ($data->{username}) passed the safety check; continuing.");
        return 1;
    }
    else {
        $logger->logcarp("Username ($data->{username}) did not pass the safety check; not continuing.");
        return 0;
    }
}

sub process_legacy_home_directory {
    my $self = shift;
    my %args = (
        data => {},
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $data = $args{data};

    if (! defined $data->{legacy_home_directory}) {
        $logger->debug("No legacy_home_directory found from command output, where user is ($data->{username}). Leaving process_legacy_home_directory.");
        return;
    }
    elsif (! exists $self->{users}{$data->{username}}) {
        $logger->debug("No user ($data->{username}) found in users table to update. Leaving process_legacy_home_directory.");
        return;
    }
    elsif (
        defined $self->{users}{$data->{username}}{legacy_home_directory} &&
        defined $data->{legacy_home_directory} &&
        $self->{users}{$data->{username}}{legacy_home_directory} eq $data->{legacy_home_directory}
    ) {
        $logger->debug("legacy_home_directory from command output is same as legacy_home_directory found in users table for user ($data->{username}). No need to update, leaving process_legacy_home_directory.");
        return;
    }

    my $sql = 'update_legacy_home_directory';
    my $sth = $sql.'_sth';

    my @params = (
        $data->{legacy_home_directory},
        $data->{username},
    );
    $logger->debug('params for update_legacy_home_directory sql:', { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of ('".join(', ', map { defined $_ ? $_ : 'undef' } @params)."')";

    my $done = -1;

    UPDATE_LEGACY_HOME_DIRECTORY:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            $logger->logcarp('DB Error Params:', { filter => \&Data::Dumper::Dumper, value => \@params });
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Updating legacy home directory, expecting update to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Updating legacy home directory, expecting update to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 or $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next UPDATE_LEGACY_HOME_DIRECTORY;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub process_passwords {
    my $self = shift;
    my %args = (
        data => {},
        map  => {},
        @_,
    );

    my $data = $args{data};
    my $map  = $args{map};

    while (my ($pw_hash_key, $pw_details) = each %{$map->{passwords}}) {
        $self->process_password(
            username   => $data->{username},
            hash       => $data->{$pw_hash_key},
            changed_at => $self->to_DateTime(timestamp => $data->{$pw_details->{changed_at}}),
            type       => $pw_details->{type},
        );
    }
}

sub process_password {
    my $self = shift;
    my %args = (
        username   => undef,
        hash       => undef,
        changed_at => undef,
        type       => 'crypt_0',
        @_,
    );
    my $logger = get_logger(__PACKAGE__);
    $logger->debug('Arguments for process_password:', { filter => \&Data::Dumper::Dumper, value => \%args });

    if (! defined $args{hash}) {
        # leave early if we don't have a password hash to work with
        $logger->debug('No password hash supplied for password processing. Exiting early.');
        return;
    }

    my $username = $args{username};
    my $type     = $args{type};

    if (exists $self->{passwords}{$username}) {
        $logger->debug("Password(s) from database for user ($username) exist.");
        if (exists $self->{passwords}{$username}{$type}) {
            $logger->debug(
                "Password of type $type from database for username ($username) "
                . "exists: \"$self->{passwords}{$username}{$type}{hash}\" with "
                . "changed_at value \""
                . ((defined $self->{passwords}{$username}{$type}{changed_at})
                ? $self->{passwords}{$username}{$type}{changed_at}->strftime("%Y-%m-%d %H:%M:%S %z")
                : 'undef')
                . q{"}
            );
            # Use smart match (~~) because some of the values could be undef and we want
            # to handle that correctly. Also instead of the preferred form of ^A || ^B
            # we use the De Morgan equivalent of ^(A && B) because the smart operator does
            # not have a built in negated form. Which would be !~~ if it existed.
            if ($self->{passwords}{$username}{$type}{hash} ~~ $args{hash} && $self->are_dates_equal(
                date_1 => $self->{passwords}{$username}{$type}{changed_at},
                date_2 => $args{changed_at},
            )) {
                $logger->debug("Password hashes and password changed at values do not differ for username ($username) and password type $type. Not updating.");
            }
            else {
                $logger->info("Password hashes or password changed at values differ for username ($username) and password type $type. Now updating.");
                $self->update_password(%args);
            }
        }
        else {
            $logger->info("No password of type $type from database for username ($username) exist. Inserting password.");
            $self->insert_password(%args);
        }
    }
    else {
        $logger->info("No passwords from database for username ($username) exist. Inserting password.");
        $self->insert_password(%args);
    }
}

sub are_dates_equal {
    my $self = shift;
    my %args = (
        date_1 => undef,
        date_2 => undef,
        @_,
    );
    my $logger = get_logger(__PACKAGE__);

    if (!defined $args{date_1} && !defined $args{date_2}) {
        return 1;
    }
    elsif (!defined $args{date_1} && defined $args{date_2}) {
        return 0;
    }
    elsif (defined $args{date_1} && !defined $args{date_2}) {
        return 0;
    }
    elsif (defined $args{date_1} && defined $args{date_2}) {
        # compare returns:
        # 0 if the DateTime objects are equal. However, we want
        # 1 if the timestamps are equal, so negate the result.
        return !DateTime->compare($args{date_1}, $args{date_2});
    }
    else {
        # Log an error. We should never reach this point.
        $logger->logconfess("Unexpected condition reached.");
    }
}

=head2
=cut
sub to_DateTime {
    my $self = shift;
    my %args = (
        timestamp => undef,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $timestamp = $args{timestamp};

    # Keep object undef if timestamp is not defined.
    return undef if !defined $timestamp;
    return undef if $timestamp eq '';

    my @timestamp_regexes = (
        qr{
            \A
            \s*
            (?<year>\d{4})
            (?<month>\d{2})
            (?<day>\d{2})
            \s*
            \z
        }xms,
        qr{
            \A
            \s*
            (?<year>\d{4})
            (?<month>\d{2})
            (?<day>\d{2})
            (?<hour>\d{2})
            (?<minute>\d{2})
            (?<second>\d{2})
            \s*
            \z
        }xms,
        qr{
            \A
            (?<year>\d{4})
            (\D)             # This is the second capture parenthesis and our date delimiter
            (?<month>\d{2})
            \2
            (?<day>\d{2})
            \s+
            (?<hour>\d{2})
            (\D)             # This is the sixth capture parenthesis and our time delimiter
            (?<minute>\d{2})
            \6
            (?<second>\d{2})
            (?<time_zone>.*?)
            \s*
            \z
        }xms,
    );

    for my $regex (@timestamp_regexes) {
        if ($timestamp =~ $regex) {
            # Our timestamp string matched the regex.
            my $dt = DateTime->new(
                year      => $+{year},
                month     => $+{month},
                day       => $+{day},
                hour      => (defined $+{hour})   ? $+{hour}   : 0,
                minute    => (defined $+{minute}) ? $+{minute} : 0,
                second    => (defined $+{second}) ? $+{second} : 0,
                time_zone => $self->normalize_timezone($+{time_zone}),
            );
            if (defined $dt) {
                return $dt;
            }
            else {
                $logger->logconfess('Unable to create DateTime object with arguments:', { filter => \&Data::Dumper::Dumper, value => \%+ });
            }
        }
    }
    $logger->logcroak("Timestamp did not match any regex and hence did not get converted to a DateTime object. Timestamp = '$timestamp'");
}

sub normalize_timezone {
    my $self      = shift;
    my $time_zone = shift;

    my $time_zone_regex = qr{
        \A
        \s*
        [-+]?
        (?:\d{2})
        \s*
        \z
    }xms;

    if (!defined $time_zone) {
        return $self->{time_zone};
    }
    elsif ($time_zone =~ $time_zone_regex) {
        return DateTime::TimeZone->new(name => $time_zone.'00');
    }
}


sub insert_password {
    my $self = shift;
    my %args = (
        username   => undef,
        hash       => undef,
        changed_at => undef,
        type       => 'crypt_0',
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    $logger->debug('%args in insert_password:', { filter => \&Data::Dumper::Dumper, value => \%args });

    my $sql = 'insert_into_passwords';
    my $sth = $sql.'_sth';

    my @params = (
        $args{username},
        $args{hash},
        $args{type},
        (defined $args{changed_at})
        ? $args{changed_at}->strftime("%Y-%m-%d %H:%M:%S %z")
        : undef,
    );
    $logger->debug('params for insert_password sql:', { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of ('".join(', ', map { defined $_ ? $_ : 'undef' } @params)."')";

    my $done = -1;

    INSERT_PASSWORD:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            $logger->logcarp('DB Error Params:', { filter => \&Data::Dumper::Dumper, value => \@params });
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Inserting into passwords, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Inserting into passwords, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 or $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next INSERT_PASSWORD;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub update_password {
    my $self = shift;
    my %args = (
        username   => undef,
        hash       => undef,
        changed_at => undef,
        type       => 'crypt_0',
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'update_passwords';
    my $sth = $sql.'_sth';

    $logger->debug('%args in update_password:', { filter => \&Data::Dumper::Dumper, value => \%args });

    my @params = (
        $args{hash},
        (defined $args{changed_at})
        ? $args{changed_at}->strftime("%Y-%m-%d %H:%M:%S %z")
        : undef,
        $args{username},
        $args{type},
    );
    $logger->debug('params for update_password sql:', { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of ('".join(', ', map { defined $_ ? $_ : 'undef' } @params)."')";

    my $done = -1;

    UPDATE_PASSWORD:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            $logger->logcarp('DB Error Params:', { filter => \&Data::Dumper::Dumper, value => \@params });
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Updating passwords, expecting update to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Updating passwords, expecting update to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 or $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next UPDATE_PASSWORD;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

1;
